#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>
#include <gsl/gsl_fit.h>
#include "fit.h"

int expd3_f (const gsl_vector * x, void *params,
        gsl_vector * f)
{
  size_t n = ((struct fitData *)params)->n;
  double *X = ((struct fitData *)params)->X;
  double *Y = ((struct fitData *)params)->Y;
  
  double A1=gsl_vector_get(x,0);
  double t1=gsl_vector_get(x,1);
  double A2=gsl_vector_get(x,2);
  double t2=gsl_vector_get(x,3);
  double A3=gsl_vector_get(x,4);
  double t3=gsl_vector_get(x,5);
  double y0=gsl_vector_get(x,6);

  size_t i;
  for (i = 0; i < n; i++)
    {
      double Yi = A1 * exp (-X[i]*t1) + A2 * exp (-X[i]*t2) + A3 * exp (-X[i]*t3) +y0;
      gsl_vector_set (f, i, Yi - Y[i]);
    }

  return GSL_SUCCESS;
}

int expd3_df (const gsl_vector * x, void *params,
         gsl_matrix * J)
{
  size_t n = ((struct fitData *)params)->n;
  double *X = ((struct fitData *)params)->X;

  double A1=gsl_vector_get(x,0);
  double l1=gsl_vector_get(x,1);
  double A2=gsl_vector_get(x,2);
  double l2=gsl_vector_get(x,3);
  double A3=gsl_vector_get(x,4);
  double l3=gsl_vector_get(x,5);

  size_t i;
  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /*       Yi = A1 * exp(-xi*l1) + A2 * exp(-xi*l2) +y0  */
      /* and the xj are the parameters (A1,l1,A2,l2,y0) */
      double t = X[i];
      double e1 = exp(-t*l1);
	  double e2 = exp(-t*l2);
	  double e3 = exp(-t*l3);
		
      gsl_matrix_set (J, i, 0, e1);
      gsl_matrix_set (J, i, 1, -t * A1 * e1);
      gsl_matrix_set (J, i, 2, e2);
	  gsl_matrix_set (J, i, 3, -t * A2 * e2);
	  gsl_matrix_set (J, i, 4, e3);
	  gsl_matrix_set (J, i, 5, -t * A3 * e3);
      gsl_matrix_set (J, i, 6, 1);
    }
  return GSL_SUCCESS;
}

int expd3_fdf (const gsl_vector * x, void *params,
          gsl_vector * f, gsl_matrix * J)
{
  expd3_f (x, params, f);
  expd3_df (x, params, J);

  return GSL_SUCCESS;
}

int expd2_f (const gsl_vector * x, void *params,
        gsl_vector * f)
{
  size_t n = ((struct fitData *)params)->n;
  double *X = ((struct fitData *)params)->X;
  double *Y = ((struct fitData *)params)->Y;
  
  double A1=gsl_vector_get(x,0);
  double t1=gsl_vector_get(x,1);
  double A2=gsl_vector_get(x,2);
  double t2=gsl_vector_get(x,3);
  double y0=gsl_vector_get(x,4);

  size_t i;
  for (i = 0; i < n; i++)
    {
      double Yi = A1 * exp (-X[i]*t1) + A2 * exp (-X[i]*t2) + y0;
      gsl_vector_set (f, i, Yi - Y[i]);
    }

  return GSL_SUCCESS;
}

int expd2_df (const gsl_vector * x, void *params,
         gsl_matrix * J)
{
  size_t n = ((struct fitData *)params)->n;
  double *X = ((struct fitData *)params)->X;

  double A1=gsl_vector_get(x,0);
  double l1=gsl_vector_get(x,1);
  double A2=gsl_vector_get(x,2);
  double l2=gsl_vector_get(x,3);

  size_t i;

  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /*       Yi = A1 * exp(-xi*l1) + A2 * exp(-xi*l2) +y0  */
      /* and the xj are the parameters (A1,l1,A2,l2,y0) */
      double t = X[i];
      double e1 = exp(-t*l1);
	  double e2 = exp(-t*l2);
		
      gsl_matrix_set (J, i, 0, e1);
      gsl_matrix_set (J, i, 1, -t * A1 * e1);
      gsl_matrix_set (J, i, 2, e2);
	  gsl_matrix_set (J, i, 3, -t * A2 * e2);
      gsl_matrix_set (J, i, 4, 1);
    }
  return GSL_SUCCESS;
}

int expd2_fdf (const gsl_vector * x, void *params,
          gsl_vector * f, gsl_matrix * J)
{
  expd2_f (x, params, f);
  expd2_df (x, params, J);

  return GSL_SUCCESS;
}

int exp_f (const gsl_vector * x, void *params,
        gsl_vector * f)
{
  size_t n = ((struct fitData *)params)->n;
  double *X = ((struct fitData *)params)->X;
  double *Y = ((struct fitData *)params)->Y;

  double A = gsl_vector_get (x, 0);
  double lambda = gsl_vector_get (x, 1);
  double b = gsl_vector_get (x, 2);

  size_t i;

  for (i = 0; i < n; i++)
    {
      double Yi = A * exp (-lambda * X[i]) + b;
	  gsl_vector_set (f, i, Yi - Y[i]);
    }

  return GSL_SUCCESS;
}

int exp_df (const gsl_vector * x, void *params,
         gsl_matrix * J)
{
  size_t n = ((struct fitData *)params)->n;
  double *X = ((struct fitData *)params)->X;

  double A = gsl_vector_get (x, 0);
  double lambda = gsl_vector_get (x, 1);

  size_t i;
  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /*       Yi = A * exp(-lambda * x[i]) + b  */
      /* and the xj are the parameters (A,lambda,b) */

      double t = X[i];
      double e = exp(-lambda * t);
	  gsl_matrix_set (J, i, 0, e);
      gsl_matrix_set (J, i, 1, -t * A * e);
      gsl_matrix_set (J, i, 2, 1);

    }
  return GSL_SUCCESS;
}

int exp_fdf (const gsl_vector * x, void *params,
          gsl_vector * f, gsl_matrix * J)
{
  exp_f (x, params, f);
  exp_df (x, params, J);

  return GSL_SUCCESS;
}

int gauss_f (const gsl_vector * x, void *params,
        gsl_vector * f)
{
  size_t n = ((struct fitData *)params)->n;
  double *X = ((struct fitData *)params)->X;
  double *Y = ((struct fitData *)params)->Y;

  double B = gsl_vector_get (x, 0);
  double A = gsl_vector_get (x, 1);
  double C = gsl_vector_get (x, 2);
  double w = gsl_vector_get (x, 3);

  size_t i;

  for (i = 0; i < n; i++)
    {
	  double diff=X[i]-C;
      double Yi = A*exp(-0.5*diff*diff/(w*w))+B;
      gsl_vector_set (f, i, Yi - Y[i]);
    }
  return GSL_SUCCESS;
}

int gauss_df (const gsl_vector * x, void *params,
         gsl_matrix * J)
{
  size_t n = ((struct fitData *)params)->n;
  double *X = ((struct fitData *)params)->X;

  double A = gsl_vector_get (x, 1);
  double C = gsl_vector_get (x, 2);
  double w = gsl_vector_get (x, 3);

  size_t i;
  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /* Yi = y=A*exp[-(Xi-xc)^2/(2*w*w)]+B  */
      /* and the xj are the parameters (B,A,C,w) */
    
	  double diff=X[i]-C;
      double e =exp(-0.5*diff*diff/(w*w));
	 
      gsl_matrix_set (J, i, 0, 1);
      gsl_matrix_set (J, i, 1, e);
      gsl_matrix_set (J, i, 2, diff*A*e/(w*w));
	  gsl_matrix_set (J, i, 3, diff*diff*A*e/(w*w*w));
    }
  return GSL_SUCCESS;
}

int gauss_fdf (const gsl_vector * x, void *params,
          gsl_vector * f, gsl_matrix * J)
{
  gauss_f (x, params, f);
  gauss_df (x, params, J);

  return GSL_SUCCESS;
}

int lorentz_f (const gsl_vector * x, void *params,
        gsl_vector * f)
{
  size_t n = ((struct fitData *)params)->n;
  double *X = ((struct fitData *)params)->X;
  double *Y = ((struct fitData *)params)->Y;

  double B = gsl_vector_get (x, 0);
  double C = gsl_vector_get (x, 2);
  double w = gsl_vector_get (x, 3);
  double aw = w*gsl_vector_get (x, 1);
  double w2 = w*w;

  size_t i;
  for (i = 0; i < n; i++)
    {
	  double diff=X[i]-C;
      double Yi = aw/(4*diff*diff+w2)+B;
      gsl_vector_set (f, i, Yi - Y[i]);
    }

  return GSL_SUCCESS;
}

int lorentz_df (const gsl_vector * x, void *params,
         gsl_matrix * J)
{
  size_t n = ((struct fitData *)params)->n;
  double *X = ((struct fitData *)params)->X;

  double A = gsl_vector_get (x, 1);
  double C = gsl_vector_get (x, 2);
  double w = gsl_vector_get (x, 3);

  size_t i;
  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /* Yi = y=A*w/[4*(Xi-xc)^2+w^2]+B  */
      /* and the xj are the parameters (B,A,C,w) */
    
	  double diff=X[i]-C;
      double num =1/(4*diff*diff+w*w);
	  double den =4*diff*diff-w*w;
	 
      gsl_matrix_set (J, i, 0, 1);
      gsl_matrix_set (J, i, 1, w*num);
      gsl_matrix_set (J, i, 2, 8*diff*A*w*num*sqrt(num));
	  gsl_matrix_set (J, i, 3, den*A*num*num);
    }
  return GSL_SUCCESS;
}

int lorentz_fdf (const gsl_vector * x, void *params,
          gsl_vector * f, gsl_matrix * J)
{
  lorentz_f (x, params, f);
  lorentz_df (x, params, J);

  return GSL_SUCCESS;
}

int print_state (size_t iter, gsl_multifit_fdfsolver * s)
{
FILE *stream;
stream=fopen("test.txt","a+");

fprintf (stream,"iter: %3u x = % 15.8f % 15.8f % 15.8f "
          "|f(x)| = %g\n",
          iter,
          gsl_vector_get (s->x, 0),
          gsl_vector_get (s->x, 1),
          gsl_vector_get (s->x, 2),
		  //gsl_vector_get (s->x, 3),
          gsl_blas_dnrm2 (s->f));
fclose(stream);
return 0;
}
